//Jose M. Pea
//CompBio
//08/01/2005

//Radial PC algorithm.

//Improvements:
//All the memory structures are static and they could be replaced by dynamic ones.
//Some arrays of integers can be replaced by arrays of binary.

//Assumptions: 
//The first line in data_file contains the names of the random variables.
//MAX_LINE_LENGTH, MAX_STATE_LENGTH.

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////includes///////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <limits.h>

#include "linalg.h"
#include "normdist.h"

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////defines//////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#define MAX_LINE_LENGTH 100000
#define MAX_STATE_LENGTH 100

#define MAX_N_CONDS 10

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////global definitions/////////////////////////////
///////////////////////////////////////////////////////////////////////////////

int n_cases, n_vars, true_net, radius, max_max_cond_size;

double k_tradeoff, critical;

char *data_file;

bool **the_true_net;

double **data;

char **names;

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////functions' prototypes///////////////////////////
///////////////////////////////////////////////////////////////////////////////

void parse_data(void);
void rpc(int target);
void pc_set(int target, int *pc, int **sep);
void pc_superset(int target, int metatarget, int *pc, int **sep);
double min_dep(int var, int target, int compulsory, int *pc, int *sep);
int next_cond_index(int n_pc, int cond_size, int *cond_index);
double compute_dep(int var, int target, int *cond);
void covariance_matrix(int n_vars, double **data, double **sigma);
int k_greedy(double *dep);
void inmoralities(int i, int *q, int **sep, bool **net);
void report(int target, clock_t clock_init, int *pc);
void prune_and_report(int target, int radius_local, clock_t clock_init, bool **net);

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////functions' body////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////

void main(int argc, char* argv[])

{int i;
 double alpha;
 char target[(int)MAX_LINE_LENGTH];

if(argc!=8)
	{printf("\n\n Usage: data_file n_cases n_vars k_tradeoff true_net(0=no,1=yes) alpha radius. \n\n"); 
	 exit(0);
	}

srand((unsigned)time(NULL));

data_file=argv[1];
n_cases=atoi(argv[2]);
n_vars=atoi(argv[3]);
k_tradeoff=atof(argv[4]);
true_net=atoi(argv[5]);
alpha=atof(argv[6]);
radius=atoi(argv[7]);

max_max_cond_size=(int)MAX_N_CONDS;
if(max_max_cond_size>n_cases-3)
	max_max_cond_size=n_cases-3;

critical=normalv(alpha/(double)2.0,(double)0.0,(double)1.0);

if(true_net==1)
	{the_true_net=new bool*[n_vars];
	 for(i=0;i<n_vars;i++)
		the_true_net[i]=new bool[n_vars];
	}

data=new double*[n_cases];
for(i=0;i<n_cases;i++)
	data[i]=new double[n_vars];

names=new char*[n_vars];
for(i=0;i<n_vars;i++)
	names[i]=new char[(int)MAX_STATE_LENGTH];

parse_data();

printf("\n\n Gene (name/all) ? ");
gets(target);

if(strcmp(target,"all")==0)
	for(i=0;i<n_vars;i++)
		rpc(i);
else
	for(i=0;i<n_vars;i++)
		if(strcmp(target,names[i])==0)
			{rpc(i);
			 break;
			}

/*
printf("\n\n Assumption:"); 
printf("\n The first line in data_file contains the names of the random variables.");
printf("\n MAX_LINE_LENGTH, MAX_STATE_LENGTH.");*/
}

/////////////////////////////////////

void parse_data(void)

{int i, j;
 float auxf;
 char aux;
 char name2[80];
 char buffer[(int)MAX_LINE_LENGTH];
 char *token;
 const char *separators=" ,\t\n\r";
 FILE *f_in;

if(true_net==1)
	{sprintf(name2,"%s.net",data_file);
	 f_in=fopen(name2,"r");

	 for(i=0;i<n_vars;i++)
		{for(j=0;j<n_vars;j++)
			{fscanf(f_in,"%c",&aux);
		     the_true_net[i][j]=(bool)atoi(&aux);
			}

		 fscanf(f_in,"%c",&aux);
		}

	 fclose(f_in);
	}

f_in=fopen(data_file,"r");

fgets(buffer,(int)MAX_LINE_LENGTH,f_in);
if((int)strlen(buffer)==(int)MAX_LINE_LENGTH-1)
	{printf("\n\n Exceeding MAX_LINE_LENGTH. \n\n");
	 exit(0);
	}
  
token=strtok(buffer,separators); 
for(j=0;j<n_vars;j++)
	{if((int)strlen(token)>(int)MAX_STATE_LENGTH-1)
		{printf("\n\n Exceeding MAX_STATE_LENGTH. \n\n");
		 exit(0);
		}
		
	 strcpy(names[j],token);
			 
	 token=strtok(0,separators);
	}

for(i=0;i<n_cases;i++)
	for(j=0;j<n_vars;j++)
		{fscanf(f_in,"%f",&auxf);
		 data[i][j]=(double)auxf;
		}

fclose(f_in);

/*
if(true_net==1)
	for(i=0;i<n_vars;i++)
		{printf("\n");
		 for(j=0;j<n_vars;j++)
			printf("%d",the_true_net[i][j]);
		}
exit(0);
*/
/*
for(j=0;j<n_vars;j++)
	printf("%s ",names[j]);
printf("\n");
for(i=0;i<n_cases;i++)
	{for(j=0;j<n_vars;j++)
		printf("%.2f ",(float)data[i][j]);
	 printf("\n");
	}
exit(0);
*/
}

/////////////////////////////////////

void rpc(int target)

{int i, j, k;
 clock_t clock_init;
 int *q, *q2;
 bool **net;
 int **sep;

//printf("\n\n rpc with target %s",names[target]);

clock_init=clock();

q=new int[n_vars];
q2=new int[n_vars];
for(i=0;i<n_vars;i++)
	{q[i]=0;
	 q2[i]=0;
	}

q[target]=1;

net=new bool*[n_vars];
for(i=0;i<n_vars;i++)
	{net[i]=new bool[n_vars];
	 for(j=0;j<n_vars;j++)
		net[i][j]=0;
	}

sep=new int*[n_vars];
for(i=0;i<n_vars;i++)
	sep[i]=new int[max_max_cond_size];

for(k=1;k<=radius+1;k++)
	{printf("\n\n Radius %d \n",k);
	 printf(" ------");
		
	 for(i=0;i<n_vars;i++)
		if(q[i]==1)
			{q[i]=-1;
	
			 pc_set(i,q2,sep);
			 report(i,clock_init,q2);

			 for(j=0;j<n_vars;j++)
				if(q2[j]==1 && q[j]==0)
					q[j]=-2;

			 for(j=0;j<n_vars;j++)
				if(q2[j]==1 && net[i][j]==0 && net[j][i]==0)
					{net[i][j]=1;
					 net[j][i]=1;
					}

			 inmoralities(i,q,sep,net);
			}

	 for(i=0;i<n_vars;i++)
		if(q[i]==-2)
			q[i]=1;

	 prune_and_report(target,k-1,clock_init,net);
	}

prune_and_report(target,radius,clock_init,net);

delete [] q;
delete [] q2;
for(i=0;i<n_vars;i++)
	{delete [] net[i];
	 delete [] sep[i];
	}
delete [] net;
delete [] sep;
}

/////////////////////////////////////

void pc_set(int target, int *pc, int **sep)

//AlgorithmPC(target). 
//It returns pc[01] and sep[][-1i].

{int i, j;
 int *pc2;
 int **sep2;

//printf("\n\n pc_set with target %s",names[target]);

pc2=new int[n_vars];

sep2=new int*[n_vars];
for(i=0;i<n_vars;i++)
	sep2[i]=new int[max_max_cond_size];

pc_superset(target,-1,pc,sep);

for(i=0;i<n_vars;i++)
	if(pc[i]==1)
		{pc_superset(i,target,pc2,sep2);

		 if(pc2[target]==0)
			{pc[i]=0;

			 for(j=0;j<max_max_cond_size;j++)
				sep[i][j]=sep2[target][j];
			}
		}

delete [] pc2;
for(i=0;i<n_vars;i++)
	delete [] sep2[i];
delete [] sep2;
}

/////////////////////////////////////

void pc_superset(int target, int metatarget, int *pc, int **sep)

//AlgorithmPCD(target).
//It returns pc[01] and sep[][-1i].

{int i, j, last_added, stop;
 int *sep2;
 double *dep, *cache_dep;

/*
if(metatarget==-1)
	printf("\n\n pc_superset with target %s",names[target]);
else
	printf("\n\n pc_superset with target %s and metatarget %s",names[target],names[metatarget]);
*/

for(i=0;i<n_vars;i++)
	for(j=0;j<max_max_cond_size;j++)
		sep[i][j]=-1;

sep2=new int[max_max_cond_size];

dep=new double[n_vars];

cache_dep=new double[n_vars];
for(i=0;i<n_vars;i++)
	cache_dep[i]=(double)0.0;

for(i=0;i<n_vars;i++)
	pc[i]=0;
pc[target]=-1;

if(metatarget!=-1)
	pc[metatarget]=1;

last_added=-1;

do
	{stop=1;

	 for(i=0;i<n_vars;i++)
		dep[i]=(double)0.0;

	 for(i=0;i<n_vars;i++)
		if(pc[i]==0)
			{dep[i]=min_dep(i,target,last_added,pc,sep2);

			 if(dep[i]==(double)0.0)
				dep[i]=cache_dep[i];
			 else
				{if(cache_dep[i]==(double)0.0 || dep[i]<cache_dep[i])
					{cache_dep[i]=dep[i];

					 if(dep[i]<=(double)(-1.0))
						for(j=0;j<max_max_cond_size;j++)
							sep[i][j]=sep2[j];
					}
				 else
					dep[i]=cache_dep[i];
				}
			}

	 for(i=0;i<n_vars;i++)
		if(dep[i]<=(double)(-1.0))
			pc[i]=-1;
		else
			if(dep[i]>=(double)1.0)
				stop=0;

	 if(stop==0)
		{last_added=k_greedy(dep);

		 pc[last_added]=1;

		 //printf("\n Adding %s with dependence %.2f",names[last_added],dep[last_added]);

		 for(i=0;i<n_vars;i++)
			dep[i]=(double)0.0;
	
		 for(i=0;i<n_vars;i++)
			if(pc[i]==1 && i!=last_added)
				{pc[i]=0; 

				 dep[i]=min_dep(i,target,last_added,pc,sep2);

				 pc[i]=1;

				 if(dep[i]<=(double)(-1.0))
					for(j=0;j<max_max_cond_size;j++)
						sep[i][j]=sep2[j];
				}

		 for(i=0;i<n_vars;i++)
			if(pc[i]==1 && i!=last_added && dep[i]<=(double)(-1.0))
				{pc[i]=-1;

				 //printf("\n Removing %s",names[i]);
				}
		}
	}
while(stop==0);

for(i=0;i<n_vars;i++)
	if(pc[i]==-1)
		pc[i]=0;

delete [] sep2;
delete [] dep;
delete [] cache_dep;
}

/////////////////////////////////////

double min_dep(int var, int target, int compulsory, int *pc, int *sep)

//It computes the minimum of the depence measure between var and target conditioned on 
//subsets of pc[x1] st compulsory is always included (-1=no compulsory). Note that only
//those subsets of pc[x1] that have a size no larger than max_cond_size are considered.
//It returns the minimum dependence [-inf,-1] for H0 (the smaller the more independent), 
//[1,+inf] for H1 (the higher the more dependent), and sep[-1i].

{int i, j, n_pc, stop, cond_size, max_cond_size;
 double dep, aux;
 int *cond, *code, *cond_index;

cond=new int[max_max_cond_size];
for(i=0;i<max_max_cond_size;i++)
	{cond[i]=-1;
	 sep[i]=-1;
	}

n_pc=0;
for(i=0;i<n_vars;i++)
	if(pc[i]==1 && i!=compulsory)
		n_pc++;

max_cond_size=max_max_cond_size;
if(compulsory!=-1)
	max_cond_size--;
if(max_cond_size>n_pc)
	max_cond_size=n_pc;

code=new int[n_pc];
j=0;
for(i=0;i<n_vars;i++)
	if(pc[i]==1 && i!=compulsory)
		{code[j]=i;

		 j++;
		}

dep=(double)0.0;
for(cond_size=0;cond_size<=max_cond_size;cond_size++)
	{cond_index=new int[cond_size];
	 for(i=0;i<cond_size;i++)
		cond_index[i]=i;

	 do
		{stop=0;
		
		 for(i=0;i<cond_size;i++)
			cond[i]=code[cond_index[i]];
		 if(compulsory!=-1)
			cond[cond_size]=compulsory;
	
		 aux=compute_dep(var,target,cond);

		 if(aux!=(double)0.0 && (dep==(double)0.0 || aux<dep))
			{dep=aux;
			 
			 if(dep<=(double)(-1.0))
				{for(i=0;i<max_max_cond_size;i++)
					sep[i]=cond[i];

				 stop=1;
				 cond_size=max_cond_size+1;
				}
			}
	 
		 if(stop==0)
			stop=next_cond_index(n_pc,cond_size,cond_index);
		}
	 while(stop==0);

	 delete [] cond_index;
	}

delete [] code;
delete [] cond;

return dep;
}

/////////////////////////////////////

int next_cond_index(int n_pc, int cond_size, int *cond_index)

//Based on the function subSetsOfSize() from the Elvira project.

{int i, j, stop;

stop=1;
for(i=cond_size-1;i>=0;i--)
	{if(cond_index[i]<n_pc+i-cond_size)
		{cond_index[i]++;
		 
		 if(i<cond_size-1)
			for(j=i+1;j<cond_size;j++)
				cond_index[j]=cond_index[j-1]+1;

		 stop=0;
		 i=-1;
		}
	}

return stop;
}

/////////////////////////////////////

double compute_dep(int var, int target, int *cond)

//It computes the depence measure between var and target conditioned on cond[-1i], where
//cond[-1i] is an array of size max_max_cond_size.
//It returns the dependence [-inf,-1] for H0 (the smaller the more independent), [+1,+inf]
//for H1 (the higher the more dependent).

//It can be speeded up by allocating memory only once in main() at the cost of memory space.

{int i, j, k, n_conds;
 double statistic;
 double **data_var_target_cond, **sigma, **sigma_var_target, **sigma_var_target_cond, **sigma_cond, **sigma_cond_var_target, **sigma_aux, **sigma_aux2;

n_conds=0;
for(i=0;i<max_max_cond_size;i++) 
	if(cond[i]!=-1)
		n_conds++;

data_var_target_cond=new double*[n_cases];
for(i=0;i<n_cases;i++)
	data_var_target_cond[i]=new double[n_conds+2];

sigma=new double*[n_conds+2];
for(i=0;i<n_conds+2;i++)
	sigma[i]=new double[n_conds+2];

sigma_var_target=new double*[2];
for(i=0;i<2;i++)
	sigma_var_target[i]=new double[2];

sigma_var_target_cond=new double*[2];
for(i=0;i<2;i++)
	sigma_var_target_cond[i]=new double[n_conds];

sigma_cond=new double*[n_conds];
for(i=0;i<n_conds;i++)
	sigma_cond[i]=new double[n_conds];

sigma_cond_var_target=new double*[n_conds];
for(i=0;i<n_conds;i++)
	sigma_cond_var_target[i]=new double[2];

sigma_aux=new double*[2];
for(i=0;i<2;i++)
	sigma_aux[i]=new double[n_conds];

sigma_aux2=new double*[2];
for(i=0;i<2;i++)
	sigma_aux2[i]=new double[2];

for(i=0;i<n_cases;i++)
	data_var_target_cond[i][0]=data[i][var];

for(i=0;i<n_cases;i++)
	data_var_target_cond[i][1]=data[i][target];

k=2;
for(j=0;j<max_max_cond_size;j++)
	if(cond[j]!=-1)
		{for(i=0;i<n_cases;i++)
			data_var_target_cond[i][k]=data[i][cond[j]];

		 k++;
		}

covariance_matrix(n_conds+2,data_var_target_cond,sigma);

for(i=0;i<2;i++)
	for(j=0;j<2;j++)	
		sigma_var_target[i][j]=sigma[i][j];

for(i=0;i<2;i++)
	for(j=2;j<n_conds+2;j++)	
		sigma_var_target_cond[i][j-2]=sigma[i][j];

for(i=2;i<n_conds+2;i++)
	for(j=2;j<n_conds+2;j++)	
		sigma_cond[i-2][j-2]=sigma[i][j];

for(i=2;i<n_conds+2;i++)
	for(j=0;j<2;j++)	
		sigma_cond_var_target[i-2][j]=sigma[i][j];

if(n_conds>0)
	{if(inverse(sigma_cond,n_conds)==0)
		{printf("\n\n No inverse. \n\n");
		 exit(0);
		}

	 matmat(sigma_var_target_cond,2,n_conds,sigma_cond,n_conds,sigma_aux);
	 matmat(sigma_aux,2,n_conds,sigma_cond_var_target,2,sigma_aux2);

	 for(i=0;i<2;i++)
		for(j=0;j<2;j++)	
			sigma_var_target[i][j]-=sigma_aux2[i][j];
	}

if(sigma_var_target[0][1]==(double)0.0)
	statistic=(double)0.0;
else
	statistic=sigma_var_target[0][1]/sqrt(sigma_var_target[0][0]*sigma_var_target[1][1]);

if(statistic==(double)1.0)
	statistic=(double)INT_MAX;
else
	statistic=(double)0.5*sqrt((double)(n_cases-n_conds-3))*log(((double)1.0+statistic)/((double)1.0-statistic));

statistic=fabs(statistic)-fabs(critical);

if(statistic>(double)0.0)
	statistic+=(double)1.0;
else
	statistic-=(double)(1.0);

for(i=0;i<n_cases;i++)
	delete [] data_var_target_cond[i];
delete [] data_var_target_cond;
for(i=0;i<n_conds+2;i++)
	delete [] sigma[i];
delete [] sigma;
for(i=0;i<2;i++)
	delete [] sigma_var_target[i];
delete [] sigma_var_target;
for(i=0;i<2;i++)
	delete [] sigma_var_target_cond[i];
delete [] sigma_var_target_cond;
for(i=0;i<n_conds;i++)
	delete [] sigma_cond[i];
delete [] sigma_cond;
for(i=0;i<n_conds;i++)
	delete [] sigma_cond_var_target[i];
delete [] sigma_cond_var_target;
for(i=0;i<2;i++)
	delete [] sigma_aux[i];
delete [] sigma_aux;
for(i=0;i<2;i++)
	delete [] sigma_aux2[i];
delete [] sigma_aux2;

return statistic;
}

/////////////////////////////////////

void covariance_matrix(int n_vars, double **data, double **sigma)

{int i, j, k;
 double *mean;

mean=new double[n_vars];

for(j=0;j<n_vars;j++)
    {mean[j]=(double)0.0;
    
	 for(i=0;i<n_cases;i++)
		mean[j]+=data[i][j];

	 mean[j]/=(double)n_cases;
    }

for(j=0;j<n_vars;j++)
	for(k=j;k<n_vars;k++)
		{sigma[j][k]=(double)0.0;
	
		 for(i=0;i<n_cases;i++)
			sigma[j][k]+=((data[i][j]-mean[j])*(data[i][k]-mean[k]));

		 sigma[j][k]/=(double)(n_cases-1);

		 sigma[k][j]=sigma[j][k];
		}

delete [] mean;
}

/////////////////////////////////////

int k_greedy(double *dep)

//k-greedy (maximization) over the nodes with dep[] >= 1.0.
//It returns the index of the maximum.

{int i, j, k, n_cands, n_cands_left, max;
 int *cand;

n_cands=0;
for(i=0;i<n_vars;i++)
	if(dep[i]>=(double)1.0)
		n_cands++;

cand=new int[n_cands];

j=0;
for(i=0;i<n_vars;i++)
	if(dep[i]>=(double)1.0)
		{cand[j]=i;

		 j++;
		}

n_cands_left=n_cands;

n_cands=(int)((double)n_cands*k_tradeoff);
if(n_cands<1)
	n_cands=1;

for(i=0;i<n_cands;i++)
	{j=rand()%n_cands_left;

	 if(i==0 || dep[cand[j]]>dep[max])
		max=cand[j];

	 for(k=j;k<n_cands_left-1;k++)
		cand[k]=cand[k+1];
	 
	 n_cands_left--;
	}

delete [] cand;

return max;
}

/////////////////////////////////////

void inmoralities(int i, int *q, int **sep, bool **net)

{int j, k, l, in_cond, n_conds;
 int *cond;

cond=new int[max_max_cond_size];
	
for(j=0;j<n_vars;j++)
	if(j!=i && q[j]==-1 && net[i][j]==0 && net[j][i]==0)
		for(k=0;k<n_vars;k++)
			if(net[i][k]==1 && net[k][i]==1 && net[j][k]==1)
				{in_cond=0;
				 n_conds=0;
				 for(l=0;l<max_max_cond_size;l++)
					{cond[l]=sep[j][l];

					 if(cond[l]==k)
						in_cond=1;

					 if(cond[l]!=-1)
						n_conds++;
					}

				 if(in_cond==0 && n_conds<max_max_cond_size)
					{cond[n_conds]=k;
					 
					 if(compute_dep(i,j,cond)>=(double)1.0)
						{net[k][i]=0;
						 net[k][j]=0;

						 //printf("\n inmorality: %s->%s<-%s",names[i],names[k],names[j]);
						}
					}
				}

delete [] cond;
}

/////////////////////////////////////

void report(int target, clock_t clock_init, int *pc)

{int i, tp, fp, fn;

printf("\n %d seconds,",(int)(clock()-clock_init)/(int)CLOCKS_PER_SEC);

if(true_net==1)
	{tp=0;
	 fp=0;
	 fn=0;

	 for(i=0;i<n_vars;i++)
		if(i!=target)
			{if(pc[i]==1)
				{if(the_true_net[i][target]==1 || the_true_net[target][i]==1)
					tp++;
				 else
					fp++;
				}
			 else
				if(the_true_net[i][target]==1 || the_true_net[target][i]==1)
					fn++;
			}

	 if(tp==0 && fp==0)
		printf(" p=%.2f",(double)(-1.0));
	 else
		printf(" p=%.2f",(double)tp/(double)(tp+fp));

	 if(tp==0 && fn==0)
		printf(" r=%.2f",(double)(-1.0));
	 else
		printf(" r=%.2f",(double)tp/(double)(tp+fn));

	 if((tp==0 && fp==0) || (tp==0 && fn==0))
		printf(" d=%.2f,",(double)(-1.0));
	 else
		printf(" d=%.2f,",sqrt(pow((double)1.0-(double)tp/(double)(tp+fp),(double)2.0)+pow((double)1.0-(double)tp/(double)(tp+fn),(double)2.0)));
	}

printf(" pc of %s:",names[target]);
for(i=0;i<n_vars;i++)
	if(pc[i]==1)
		printf(" %s,",names[i]);
}

/////////////////////////////////////

void prune_and_report(int target, int radius_local, clock_t clock_init, bool **net)

{int i, j, k, tp, fp, fn;
 int *in_radius;
 bool **the_true_net_local, **net_local;

net_local=new bool*[n_vars];
for(i=0;i<n_vars;i++)
	{net_local[i]=new bool[n_vars];
	 for(j=0;j<n_vars;j++)
		net_local[i][j]=net[i][j];
	}

in_radius=new int[n_vars];
for(i=0;i<n_vars;i++)
	in_radius[i]=0;
	
in_radius[target]=1;
	
for(k=1;k<=radius_local;k++)
	{for(i=0;i<n_vars;i++)
		if(in_radius[i]==1)
			{in_radius[i]=-1;

			 for(j=0;j<n_vars;j++)
				if((net_local[i][j]==1 || net_local[j][i]==1) && in_radius[j]==0)
					in_radius[j]=-2;
			}

	 for(i=0;i<n_vars;i++)
		if(in_radius[i]==-2)
			in_radius[i]=1;
	}

for(i=0;i<n_vars;i++)
	if(in_radius[i]==-1)
		in_radius[i]=1;

for(i=0;i<n_vars-1;i++)
	for(j=i+1;j<n_vars;j++)
		if((net_local[i][j]==1 || net_local[j][i]==1) && (in_radius[i]==0 || in_radius[j]==0))
			{net_local[i][j]=0;
			 net_local[j][i]=0;
			}

if(true_net==1)
	{the_true_net_local=new bool*[n_vars];
	 for(i=0;i<n_vars;i++)
		{the_true_net_local[i]=new bool[n_vars];
		 for(j=0;j<n_vars;j++)
			the_true_net_local[i][j]=the_true_net[i][j];
		}
	
	 for(i=0;i<n_vars;i++)
		in_radius[i]=0;
		
	 in_radius[target]=1;
		
	 for(k=1;k<=radius_local;k++)
		{for(i=0;i<n_vars;i++)
			if(in_radius[i]==1)
				{in_radius[i]=-1;

				 for(j=0;j<n_vars;j++)
					if((the_true_net_local[i][j]==1 || the_true_net_local[j][i]==1) && in_radius[j]==0)
						in_radius[j]=-2;
				}

		 for(i=0;i<n_vars;i++)
			if(in_radius[i]==-2)
				in_radius[i]=1;
		}

	 for(i=0;i<n_vars;i++)
		if(in_radius[i]==-1)
			in_radius[i]=1;

	 for(i=0;i<n_vars-1;i++)
		for(j=i+1;j<n_vars;j++)
			if((the_true_net_local[i][j]==1 || the_true_net_local[j][i]==1) && (in_radius[i]==0 || in_radius[j]==0))
				{the_true_net_local[i][j]=0;
				 the_true_net_local[j][i]=0;
				}

	 printf("\n\n Edges\n");
	 printf(" -----");

	 printf("\n %d seconds,",(int)(clock()-clock_init)/(int)CLOCKS_PER_SEC);

	 tp=0;
	 fp=0;
	 fn=0;

	 for(i=0;i<n_vars-1;i++)
		for(j=i+1;j<n_vars;j++)
			if(net_local[i][j]==1 || net_local[j][i]==1)
				{if(the_true_net_local[i][j]==1 || the_true_net_local[j][i]==1)
					tp++;
				 else
					fp++;
				}
			else
				if(the_true_net_local[i][j]==1 || the_true_net_local[j][i]==1)
					fn++;

	 if(tp==0 && fp==0)
		printf(" p=%.2f",(double)(-1.0));
	 else
		printf(" p=%.2f",(double)tp/(double)(tp+fp));

	 if(tp==0 && fn==0)
		printf(" r=%.2f",(double)(-1.0));
	 else
		printf(" r=%.2f",(double)tp/(double)(tp+fn));

	 if((tp==0 && fp==0) || (tp==0 && fn==0))
		printf(" d=%.2f,",(double)(-1.0));
	 else
		printf(" d=%.2f,",sqrt(pow((double)1.0-(double)tp/(double)(tp+fp),(double)2.0)+pow((double)1.0-(double)tp/(double)(tp+fn),(double)2.0)));

	 printf("\n\n Inmoralities\n");
	 printf(" ------------");

	 printf("\n %d seconds,",(int)(clock()-clock_init)/(int)CLOCKS_PER_SEC);

	 tp=0;
	 fp=0;
	 fn=0;

	 for(i=0;i<n_vars-1;i++)
		for(j=i+1;j<n_vars;j++)
			for(k=0;k<n_vars;k++)
				if(net_local[i][k]==1 && net_local[k][i]==0 && net_local[j][k]==1 && net_local[k][j]==0 && net_local[i][j]==0 && net_local[j][i]==0)
					{if(the_true_net_local[i][k]==1 && the_true_net_local[k][i]==0 && the_true_net_local[j][k]==1 && the_true_net_local[k][j]==0 && the_true_net_local[i][j]==0 && the_true_net_local[j][i]==0)
						tp++;
					 else
						fp++;
					}
				else
					if(the_true_net_local[i][k]==1 && the_true_net_local[k][i]==0 && the_true_net_local[j][k]==1 && the_true_net_local[k][j]==0 && the_true_net_local[i][j]==0 && the_true_net_local[j][i]==0)
						fn++;

	 if(tp==0 && fp==0)
		printf(" p=%.2f",(double)(-1.0));
	 else
		printf(" p=%.2f",(double)tp/(double)(tp+fp));

	 if(tp==0 && fn==0)
		printf(" r=%.2f",(double)(-1.0));
	 else
		printf(" r=%.2f",(double)tp/(double)(tp+fn));

	 if((tp==0 && fp==0) || (tp==0 && fn==0))
		printf(" d=%.2f,",(double)(-1.0));
	 else
		printf(" d=%.2f,",sqrt(pow((double)1.0-(double)tp/(double)(tp+fp),(double)2.0)+pow((double)1.0-(double)tp/(double)(tp+fn),(double)2.0)));
	
	 for(i=0;i<n_vars;i++)
		delete [] the_true_net_local[i];
	 delete [] the_true_net_local;
	}

printf("\n\n Net\n");
printf(" ---");

printf("\n\n digraph G { size=\"7.50,10\" { ");

for(i=0;i<n_vars;i++)
	for(j=0;j<n_vars;j++)
		if(net_local[i][j]==1 || net_local[j][i]==1)
			{printf("%s ",names[i]);
			 break;
			}

printf(";");

for(i=0;i<n_vars;i++)
	for(j=0;j<n_vars;j++)
		if(net_local[i][j]==1 && net_local[j][i]==0)
			printf("\n %s -> %s;",names[i],names[j]);
		else
			if(net_local[i][j]==1 && net_local[j][i]==1 && i<j)
				printf("\n %s -> %s [arrowhead=none];",names[i],names[j]);

printf(" } }");

delete [] in_radius;
for(i=0;i<n_vars;i++)
	delete [] net_local[i];
delete [] net_local;
}