//Jose M. Pea
//CompBio
//08/01/2005

//Radial PC algorithm.

//Improvements:
//All the memory structures are static and they could be replaced by dynamic ones.
//Some arrays of integers can be replaced by arrays of binary.

//Assumptions: 
//The first line in data_file contains the names of the random variables.
//MAX_LINE_LENGTH, MAX_N_STATES, MAX_STATE_LENGTH.

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////includes///////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <time.h>

#include "chisq.h"

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////defines//////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#define MAX_LINE_LENGTH 100000
#define MAX_N_STATES 10
#define MAX_STATE_LENGTH 100

#define N_CASES_PER_DF 5

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////global definitions/////////////////////////////
///////////////////////////////////////////////////////////////////////////////

int n_cases, n_vars, true_net, radius, max_max_cond_size;

double k_tradeoff, alpha;

int *n_states;

char *data_file;

bool **the_true_net;

int **data;

char **names;

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////functions' prototypes///////////////////////////
///////////////////////////////////////////////////////////////////////////////

void parse_data(void);
int state_index(int var, char *state, char ***states);
void rpc(int target);
void pc_set(int target, int *pc, int **sep);
void pc_superset(int target, int metatarget, int *pc, int **sep);
double min_dep(int var, int target, int compulsory, int *pc, int *sep);
int next_cond_index(int n_pc, int cond_size, int *cond_index);
double compute_dep(int var, int target, int *cond);
int k_greedy(double *dep);
void inmoralities(int i, int *q, int **sep, bool **net);
void report(int target, clock_t clock_init, int *pc);
void prune_and_report(int target, int radius_local, clock_t clock_init, bool **net);


///////////////////////////////////////////////////////////////////////////////
////////////////////////////////functions' body////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////

void main(int argc, char* argv[])

{int i, target;

if(argc!=9)
	{printf("\n\n Usage: data_file n_cases n_vars target(-1=all) k_tradeoff true_net(0=no,1=yes) alpha radius. \n\n"); 
	 exit(0);
	}

srand((unsigned)time(NULL));

data_file=argv[1];
n_cases=atoi(argv[2]);
n_vars=atoi(argv[3]);
target=atoi(argv[4]);
k_tradeoff=atof(argv[5]);
true_net=atoi(argv[6]);
alpha=atof(argv[7]);
radius=atoi(argv[8]);

n_states=new int[n_vars];

if(true_net==1)
	{the_true_net=new bool*[n_vars];
	 for(i=0;i<n_vars;i++)
		the_true_net[i]=new bool[n_vars];
	}

data=new int*[n_cases];
for(i=0;i<n_cases;i++)
	data[i]=new int[n_vars];

names=new char*[n_vars];
for(i=0;i<n_vars;i++)
	names[i]=new char[(int)MAX_STATE_LENGTH];

parse_data();

if(target==-1)
	for(i=0;i<n_vars;i++)	
		rpc(i);
else
	rpc(target);
/*
printf("\n\n Assumption:"); 
printf("\n The first line in data_file contains the names of the random variables.");
printf("\n MAX_LINE_LENGTH, MAX_N_STATES, MAX_STATE_LENGTH.");*/
}

/////////////////////////////////////

void parse_data(void)

{int i, j, min_n_states;
 char aux;
 char name2[80];
 char buffer[(int)MAX_LINE_LENGTH];
 char *token;
 const char *separators=" ,\t\n\r";
 FILE *f_in;
 char ***states;

if(true_net==1)
	{sprintf(name2,"%s.net",data_file);
	 f_in=fopen(name2,"r");

	 for(i=0;i<n_vars;i++)
		{for(j=0;j<n_vars;j++)
			{fscanf(f_in,"%c",&aux);
		     the_true_net[i][j]=(bool)atoi(&aux);
			}

		 fscanf(f_in,"%c",&aux);
		}

	 fclose(f_in);
	}

states=new char**[n_vars];
for(i=0;i<n_vars;i++)
	{states[i]=new char*[(int)MAX_N_STATES];
	 for(j=0;j<(int)MAX_N_STATES;j++)
		states[i][j]=new char[(int)MAX_STATE_LENGTH];
	}

for(i=0;i<n_vars;i++)
	n_states[i]=0;

f_in=fopen(data_file,"r");

for(i=-1;i<n_cases;i++)
	{fgets(buffer,(int)MAX_LINE_LENGTH,f_in);
	 if((int)strlen(buffer)==(int)MAX_LINE_LENGTH-1)
		{printf("\n\n Exceeding MAX_LINE_LENGTH. \n\n");
		 exit(0);
		}
  
	 token=strtok(buffer,separators); 
	 for(j=0;j<n_vars;j++)
		{if((int)strlen(token)>(int)MAX_STATE_LENGTH-1)
			{printf("\n\n Exceeding MAX_STATE_LENGTH. \n\n");
			 exit(0);
			}

		 if(i==-1)
			strcpy(names[j],token);
		 else
			data[i][j]=state_index(j,token,states);
			 
		 token=strtok(0,separators);
		}
	}

fclose(f_in);

min_n_states=0;
for(i=0;i<n_vars;i++)
	if(n_states[i]>1 && (min_n_states==0 || n_states[i]<min_n_states))
		min_n_states=n_states[i];

if(min_n_states==0)
	{printf("\n\n Only one state for all the variables. \n\n");
	 exit(0);
	}

max_max_cond_size=(int)(log((double)n_cases/(double)((int)N_CASES_PER_DF*(min_n_states-1)*(min_n_states-1)))/log((double)min_n_states)); //min_n_states may not be representative.

for(i=0;i<n_vars;i++)
	{for(j=0;j<(int)MAX_N_STATES;j++)
		delete [] states[i][j];
	 delete [] states[i];
	}
delete [] states;

/*
if(true_net==1)
	for(i=0;i<n_vars;i++)
		{printf("\n");
		 for(j=0;j<n_vars;j++)
			printf("%d",the_true_net[i][j]);
		}
exit(0);
*/
/*
for(i=0;i<n_cases;i++)
	{for(j=0;j<n_vars;j++)
		printf("%d ",data[i][j]);
	 printf("\n");
	}
exit(0);
*/
/*
for(i=0;i<n_vars;i++)
	printf("%d ",n_states[i]);
printf("\n");
exit(0);
*/
}

/////////////////////////////////////

int state_index(int var, char *state, char ***states)

{int i;

for(i=0;i<n_states[var];i++)
	if(strcmp(states[var][i],state)==0)
		return i;

if(i>(int)MAX_N_STATES-1)
	{printf("\n\n Exceeding MAX_N_STATES. \n\n");
	 exit(0);
	}

strcpy(states[var][i],state);
n_states[var]++;

return i;
}

/////////////////////////////////////

void rpc(int target)

{int i, j, k;
 clock_t clock_init;
 int *q, *q2;
 bool **net;
 int **sep;

//printf("\n\n rpc with target %d",target);

clock_init=clock();

q=new int[n_vars];
q2=new int[n_vars];
for(i=0;i<n_vars;i++)
	{q[i]=0;
	 q2[i]=0;
	}

q[target]=1;

net=new bool*[n_vars];
for(i=0;i<n_vars;i++)
	{net[i]=new bool[n_vars];
	 for(j=0;j<n_vars;j++)
		net[i][j]=0;
	}

sep=new int*[n_vars];
for(i=0;i<n_vars;i++)
	sep[i]=new int[max_max_cond_size];

for(k=1;k<=radius+1;k++)
	{printf("\n\n Radius %d \n",k);
	 printf(" ------");
		
	 for(i=0;i<n_vars;i++)
		if(q[i]==1)
			{q[i]=-1;
	
			 pc_set(i,q2,sep);
			 report(i,clock_init,q2);

			 for(j=0;j<n_vars;j++)
				if(q2[j]==1 && q[j]==0)
					q[j]=-2;

			 for(j=0;j<n_vars;j++)
				if(q2[j]==1 && net[i][j]==0 && net[j][i]==0)
					{net[i][j]=1;
					 net[j][i]=1;
					}

			 inmoralities(i,q,sep,net);
			}

	 for(i=0;i<n_vars;i++)
		if(q[i]==-2)
			q[i]=1;

	 prune_and_report(target,k-1,clock_init,net);
	}

prune_and_report(target,radius,clock_init,net);

delete [] q;
delete [] q2;
for(i=0;i<n_vars;i++)
	{delete [] net[i];
	 delete [] sep[i];
	}
delete [] net;
delete [] sep;
}

/////////////////////////////////////

void pc_set(int target, int *pc, int **sep)

//AlgorithmPC(target). 
//It returns pc[01] and sep[][-1i].

{int i, j;
 int *pc2;
 int **sep2;

//printf("\n\n pc_set with target %d",target);

pc2=new int[n_vars];

sep2=new int*[n_vars];
for(i=0;i<n_vars;i++)
	sep2[i]=new int[max_max_cond_size];

pc_superset(target,-1,pc,sep);

for(i=0;i<n_vars;i++)
	if(pc[i]==1)
		{pc_superset(i,target,pc2,sep2);

		 if(pc2[target]==0)
			{pc[i]=0;

			 for(j=0;j<max_max_cond_size;j++)
				sep[i][j]=sep2[target][j];
			}
		}

delete [] pc2;
for(i=0;i<n_vars;i++)
	delete [] sep2[i];
delete [] sep2;
}

/////////////////////////////////////

void pc_superset(int target, int metatarget, int *pc, int **sep)

//AlgorithmPCD(target).
//It returns pc[01] and sep[][-1i].

{int i, j, last_added, stop;
 int *sep2;
 double *dep, *cache_dep;

//printf("\n\n pc_superset with target %d and metatarget %d",target,metatarget);

for(i=0;i<n_vars;i++)
	for(j=0;j<max_max_cond_size;j++)
		sep[i][j]=-1;

sep2=new int[max_max_cond_size];

dep=new double[n_vars];

cache_dep=new double[n_vars];
for(i=0;i<n_vars;i++)
	cache_dep[i]=(double)0.0;

for(i=0;i<n_vars;i++)
	if(n_states[i]>1 && n_states[target]>1)
		pc[i]=0;
	else
		pc[i]=-1;
pc[target]=-1;

if(metatarget!=-1)
	pc[metatarget]=1;

last_added=-1;

do
	{stop=1;

	 for(i=0;i<n_vars;i++)
		dep[i]=(double)0.0;

	 for(i=0;i<n_vars;i++)
		if(pc[i]==0)
			{dep[i]=min_dep(i,target,last_added,pc,sep2);

			 if(dep[i]==(double)0.0)
				dep[i]=cache_dep[i];
			 else
				{if(cache_dep[i]==(double)0.0 || dep[i]<cache_dep[i])
					{cache_dep[i]=dep[i];

					 if(dep[i]<=(double)(-1.0))
						for(j=0;j<max_max_cond_size;j++)
							sep[i][j]=sep2[j];
					}
				 else
					dep[i]=cache_dep[i];
				}
			}

	 for(i=0;i<n_vars;i++)
		if(dep[i]<=(double)(-1.0))
			pc[i]=-1;
		else
			if(dep[i]>=(double)1.0)
				stop=0;

	 if(stop==0)
		{last_added=k_greedy(dep);

		 pc[last_added]=1;

		 //printf("\n Adding %d with dependence %.2f",last_added,dep[last_added]);

		 for(i=0;i<n_vars;i++)
			dep[i]=(double)0.0;
	
		 for(i=0;i<n_vars;i++)
			if(pc[i]==1 && i!=last_added)
				{pc[i]=0; 

				 dep[i]=min_dep(i,target,last_added,pc,sep2);

				 pc[i]=1;

				 if(dep[i]<=(double)(-1.0))
					for(j=0;j<max_max_cond_size;j++)
						sep[i][j]=sep2[j];
				}

		 for(i=0;i<n_vars;i++)
			if(pc[i]==1 && i!=last_added && dep[i]<=(double)(-1.0))
				{pc[i]=-1;

				 //printf("\n Removing %d",i);
				}
		}
	}
while(stop==0);

for(i=0;i<n_vars;i++)
	if(pc[i]==-1)
		pc[i]=0;

delete [] sep2;
delete [] dep;
delete [] cache_dep;
}

/////////////////////////////////////

double min_dep(int var, int target, int compulsory, int *pc, int *sep)

//It computes the minimum of the depence measure between var and target conditioned on 
//subsets of pc[x1] st compulsory is always included (-1=no compulsory). Note that only
//those subsets of pc[x1] that have a size no larger than max_cond_size are considered.
//It returns the minimum dependence [-inf,-1] for H0 (the smaller the more independent), 
//[1,+inf] for H1 (the higher the more dependent), or 0 for not enough data (see N_CASES_PER_DF),
//and sep[-1i].

{int i, j, n_pc, stop, cond_size, max_cond_size;
 double dep, aux;
 int *cond, *code, *cond_index;

cond=new int[max_max_cond_size];
for(i=0;i<max_max_cond_size;i++)
	{cond[i]=-1;
	 sep[i]=-1;
	}

n_pc=0;
for(i=0;i<n_vars;i++)
	if(pc[i]==1 && i!=compulsory)
		n_pc++;

max_cond_size=max_max_cond_size;
if(compulsory!=-1)
	max_cond_size--;
if(max_cond_size>n_pc)
	max_cond_size=n_pc;

code=new int[n_pc];
j=0;
for(i=0;i<n_vars;i++)
	if(pc[i]==1 && i!=compulsory)
		{code[j]=i;

		 j++;
		}

dep=(double)0.0;
for(cond_size=0;cond_size<=max_cond_size;cond_size++)
	{cond_index=new int[cond_size];
	 for(i=0;i<cond_size;i++)
		cond_index[i]=i;

	 do
		{stop=0;
		
		 for(i=0;i<cond_size;i++)
			cond[i]=code[cond_index[i]];
		 if(compulsory!=-1)
			cond[cond_size]=compulsory;
	
		 aux=compute_dep(var,target,cond);

		 if(aux!=(double)0.0 && (dep==(double)0.0 || aux<dep))
			{dep=aux;
			 
			 if(dep<=(double)(-1.0))
				{for(i=0;i<max_max_cond_size;i++)
					sep[i]=cond[i];

				 stop=1;
				 cond_size=max_cond_size+1;
				}
			}
	 
		 if(stop==0)
			stop=next_cond_index(n_pc,cond_size,cond_index);
		}
	 while(stop==0);

	 delete [] cond_index;
	}

delete [] code;
delete [] cond;

return dep;
}

/////////////////////////////////////

int next_cond_index(int n_pc, int cond_size, int *cond_index)

//Based on the function subSetsOfSize() from the Elvira project.

{int i, j, stop;

stop=1;
for(i=cond_size-1;i>=0;i--)
	{if(cond_index[i]<n_pc+i-cond_size)
		{cond_index[i]++;
		 
		 if(i<cond_size-1)
			for(j=i+1;j<cond_size;j++)
				cond_index[j]=cond_index[j-1]+1;

		 stop=0;
		 i=-1;
		}
	}

return stop;
}

/////////////////////////////////////

double compute_dep(int var, int target, int *cond)

//It computes the depence measure between var and target conditioned on cond[-1i], where
//cond[-1i] is an array of size max_max_cond_size.
//It returns the dependence [-inf,-1] for H0 (the smaller the more independent), [+1,+inf]
//for H1 (the higher the more dependent), or 0 for not enough data (see N_CASES_PER_DF).

//It can be speeded up by allocating memory only once in main() at the cost of memory space, 
//and by caching the critical value of the test as a function of the degrees of freedom.
//Only ss[][][] is strictly necessary. The rest of the arrays are used to gain speed at the 
//cost of memory space.

{int i, j, k, n_cond_states, cond_state, df, df_target, df_var;
 double statistic, dep, aux;
 int *ss_cond;
 int **ss_target, **ss_var;
 int ***ss;

n_cond_states=1;
for(i=0;i<max_max_cond_size;i++) 
	if(cond[i]!=-1)
		n_cond_states=n_cond_states*n_states[cond[i]];

if(n_cond_states*(n_states[target]-1)*(n_states[var]-1)*(int)N_CASES_PER_DF>n_cases)
	return (double)0.0;

ss_cond=new int[n_cond_states];
for(i=0;i<n_cond_states;i++)
	ss_cond[i]=0;

ss_target=new int*[n_cond_states];
for(i=0;i<n_cond_states;i++)
	{ss_target[i]=new int[n_states[target]];
	 for(j=0;j<n_states[target];j++)
		ss_target[i][j]=0;
	}

ss_var=new int*[n_cond_states];
for(i=0;i<n_cond_states;i++)
	{ss_var[i]=new int[n_states[var]];
	 for(j=0;j<n_states[var];j++)
		ss_var[i][j]=0;
	}

ss=new int**[n_cond_states];
for(i=0;i<n_cond_states;i++)
	{ss[i]=new int*[n_states[target]];
	 for(j=0;j<n_states[target];j++)
		{ss[i][j]=new int[n_states[var]];
		 for(k=0;k<n_states[var];k++)
			 ss[i][j][k]=0;
		}
	}

for(i=0;i<n_cases;i++)
	{cond_state=0;
	 for(j=0;j<max_max_cond_size;j++)
		if(cond[j]!=-1)
			cond_state=cond_state*n_states[cond[j]]+(int)data[i][cond[j]];

	 ss_cond[cond_state]++;
	 ss_target[cond_state][(int)data[i][target]]++;
	 ss_var[cond_state][(int)data[i][var]]++;
	 ss[cond_state][(int)data[i][target]][(int)data[i][var]]++;
	}

//X^2 statistic.
/*
statistic=(double)0.0;
for(i=0;i<n_cond_states;i++)
	if(ss_cond[i]>0)
		for(j=0;j<n_states[target];j++)
			if(ss_target[i][j]>0)
				for(k=0;k<n_states[var];k++)
					if(ss_var[i][k]>0)
						{aux=(double)(ss_target[i][j]*ss_var[i][k])/(double)ss_cond[i];
						 statistic=statistic+pow((double)ss[i][j][k]-aux,(double)2.0)/aux;
						}
*/

//G^2 statistic based on mutual information.
/*
statistic=(double)0.0;
for(i=0;i<n_cond_states;i++)
	if(ss_cond[i]>0)
		for(j=0;j<n_states[target];j++)
			if(ss_target[i][j]>0)
				for(k=0;k<n_states[var];k++)
					if(ss[i][j][k]>0)
						statistic=statistic+(double)ss[i][j][k]*(log10((double)ss[i][j][k])-
						log10((double)ss_target[i][j])-log10((double)ss_var[i][k])+
						log10((double)ss_cond[i]))/log10((double)2.0);
statistic=statistic*(double)2.0;
*/

//G^2 statistic based on observed and expected frequencies.

statistic=(double)0.0;
for(i=0;i<n_cond_states;i++)
	if(ss_cond[i]>0)
		for(j=0;j<n_states[target];j++)
			if(ss_target[i][j]>0)
				for(k=0;k<n_states[var];k++)
					if(ss[i][j][k]>0)
						{aux=(double)(ss_target[i][j]*ss_var[i][k])/(double)ss_cond[i];
						 statistic=statistic+(double)ss[i][j][k]*(log((double)ss[i][j][k])-log(aux));
						}		
statistic=statistic*(double)2.0;


//Reduced df due to zero entries. This does not make sense with the X^2 statistic.
/*
df=n_cond_states*(n_states[target]-1)*(n_states[var]-1);
for(i=0;i<n_cond_states;i++)
	for(j=0;j<n_states[target];j++)
		for(k=0;k<n_states[var];k++)
			if(ss[i][j][k]==0)
				df--;
*/

//Reduced df due to zero marginals.

df=0;
for(i=0;i<n_cond_states;i++)
	if(ss_cond[i]>0)
		{df_target=0;
		 for(j=0;j<n_states[target];j++)
			if(ss_target[i][j]>0)
				df_target++;

		 df_var=0;
		 for(k=0;k<n_states[var];k++)
			 if(ss_var[i][k]>0)
				df_var++;

		 df=df+(df_target-1)*(df_var-1);
		}


delete [] ss_cond;
for(i=0;i<n_cond_states;i++)
	{delete [] ss_target[i];
	 delete [] ss_var[i];
	 for(j=0;j<n_states[target];j++)
		delete [] ss[i][j];
	 delete [] ss[i];
	}
delete [] ss_target;
delete [] ss_var;
delete [] ss;

if(statistic<(double)0.0) //statistic sometimes takes values -0.0. Due to loss of precision ?
	statistic=(double)0.0;

if(df<=0) //Naive ?
	df=1;

dep=gammq((double)0.5*(double)df,(double)0.5*statistic);

if(dep<(double)0.0) //Just in case.
	dep=(double)0.0;
else
	if(dep>(double)1.0)
		dep=(double)1.0;

if(dep<=alpha)
	{dep=(double)2.0-dep;
	
	 if(dep==(double)2.0)
		dep=(double)2.0+statistic/(double)df;
	
	 return dep;
	}
else
	{dep=(double)(-1.0)-dep;
	
	 if(dep==(double)(-2.0))
		dep=(double)(-2.0)-statistic/(double)df;
	
	 return dep;
	}
}

/////////////////////////////////////

int k_greedy(double *dep)

//k-greedy (maximization) over the nodes with dep[] >= 1.0.
//It returns the index of the maximum.

{int i, j, k, n_cands, n_cands_left, max;
 int *cand;

n_cands=0;
for(i=0;i<n_vars;i++)
	if(dep[i]>=(double)1.0)
		n_cands++;

cand=new int[n_cands];

j=0;
for(i=0;i<n_vars;i++)
	if(dep[i]>=(double)1.0)
		{cand[j]=i;

		 j++;
		}

n_cands_left=n_cands;

n_cands=(int)((double)n_cands*k_tradeoff);
if(n_cands<1)
	n_cands=1;

for(i=0;i<n_cands;i++)
	{j=rand()%n_cands_left;

	 if(i==0 || dep[cand[j]]>dep[max])
		max=cand[j];

	 for(k=j;k<n_cands_left-1;k++)
		cand[k]=cand[k+1];
	 
	 n_cands_left--;
	}

delete [] cand;

return max;
}

/////////////////////////////////////

void inmoralities(int i, int *q, int **sep, bool **net)

{int j, k, l, in_cond, n_conds;
 int *cond;

cond=new int[max_max_cond_size];
	
for(j=0;j<n_vars;j++)
	if(j!=i && q[j]==-1 && net[i][j]==0 && net[j][i]==0)
		for(k=0;k<n_vars;k++)
			if(net[i][k]==1 && net[k][i]==1 && net[j][k]==1)
				{in_cond=0;
				 n_conds=0;
				 for(l=0;l<max_max_cond_size;l++)
					{cond[l]=sep[j][l];

					 if(cond[l]==k)
						in_cond=1;

					 if(cond[l]!=-1)
						n_conds++;
					}

				 if(in_cond==0 && n_conds<max_max_cond_size)
					{cond[n_conds]=k;
					 
					 if(compute_dep(i,j,cond)>=(double)1.0)
						{net[k][i]=0;
						 net[k][j]=0;

						 //printf("\n inmorality: %s->%s<-%s",names[i],names[k],names[j]);
						}
					}
				}

delete [] cond;
}

/////////////////////////////////////

void report(int target, clock_t clock_init, int *pc)

{int i, tp, fp, fn;

printf("\n %d seconds,",(int)(clock()-clock_init)/(int)CLOCKS_PER_SEC);

if(true_net==1)
	{tp=0;
	 fp=0;
	 fn=0;

	 for(i=0;i<n_vars;i++)
		if(i!=target)
			{if(pc[i]==1)
				{if(the_true_net[i][target]==1 || the_true_net[target][i]==1)
					tp++;
				 else
					fp++;
				}
			 else
				if(the_true_net[i][target]==1 || the_true_net[target][i]==1)
					fn++;
			}

	 if(tp==0 && fp==0)
		printf(" p=%.2f",(double)(-1.0));
	 else
		printf(" p=%.2f",(double)tp/(double)(tp+fp));

	 if(tp==0 && fn==0)
		printf(" r=%.2f",(double)(-1.0));
	 else
		printf(" r=%.2f",(double)tp/(double)(tp+fn));

	 if((tp==0 && fp==0) || (tp==0 && fn==0))
		printf(" d=%.2f,",(double)(-1.0));
	 else
		printf(" d=%.2f,",sqrt(pow((double)1.0-(double)tp/(double)(tp+fp),(double)2.0)+pow((double)1.0-(double)tp/(double)(tp+fn),(double)2.0)));
	}

printf(" pc of %s:",names[target]);
for(i=0;i<n_vars;i++)
	if(pc[i]==1)
		printf(" %s,",names[i]);
}

/////////////////////////////////////

void prune_and_report(int target, int radius_local, clock_t clock_init, bool **net)

{int i, j, k, tp, fp, fn;
 int *in_radius;
 bool **the_true_net_local, **net_local;

net_local=new bool*[n_vars];
for(i=0;i<n_vars;i++)
	{net_local[i]=new bool[n_vars];
	 for(j=0;j<n_vars;j++)
		net_local[i][j]=net[i][j];
	}

in_radius=new int[n_vars];
for(i=0;i<n_vars;i++)
	in_radius[i]=0;
	
in_radius[target]=1;
	
for(k=1;k<=radius_local;k++)
	{for(i=0;i<n_vars;i++)
		if(in_radius[i]==1)
			{in_radius[i]=-1;

			 for(j=0;j<n_vars;j++)
				if((net_local[i][j]==1 || net_local[j][i]==1) && in_radius[j]==0)
					in_radius[j]=-2;
			}

	 for(i=0;i<n_vars;i++)
		if(in_radius[i]==-2)
			in_radius[i]=1;
	}

for(i=0;i<n_vars;i++)
	if(in_radius[i]==-1)
		in_radius[i]=1;

for(i=0;i<n_vars-1;i++)
	for(j=i+1;j<n_vars;j++)
		if((net_local[i][j]==1 || net_local[j][i]==1) && (in_radius[i]==0 || in_radius[j]==0))
			{net_local[i][j]=0;
			 net_local[j][i]=0;
			}

if(true_net==1)
	{the_true_net_local=new bool*[n_vars];
	 for(i=0;i<n_vars;i++)
		{the_true_net_local[i]=new bool[n_vars];
		 for(j=0;j<n_vars;j++)
			the_true_net_local[i][j]=the_true_net[i][j];
		}
	
	 for(i=0;i<n_vars;i++)
		in_radius[i]=0;
		
	 in_radius[target]=1;
		
	 for(k=1;k<=radius_local;k++)
		{for(i=0;i<n_vars;i++)
			if(in_radius[i]==1)
				{in_radius[i]=-1;

				 for(j=0;j<n_vars;j++)
					if((the_true_net_local[i][j]==1 || the_true_net_local[j][i]==1) && in_radius[j]==0)
						in_radius[j]=-2;
				}

		 for(i=0;i<n_vars;i++)
			if(in_radius[i]==-2)
				in_radius[i]=1;
		}

	 for(i=0;i<n_vars;i++)
		if(in_radius[i]==-1)
			in_radius[i]=1;

	 for(i=0;i<n_vars-1;i++)
		for(j=i+1;j<n_vars;j++)
			if((the_true_net_local[i][j]==1 || the_true_net_local[j][i]==1) && (in_radius[i]==0 || in_radius[j]==0))
				{the_true_net_local[i][j]=0;
				 the_true_net_local[j][i]=0;
				}

	 printf("\n\n Edges\n");
	 printf(" -----");

	 printf("\n %d seconds,",(int)(clock()-clock_init)/(int)CLOCKS_PER_SEC);

	 tp=0;
	 fp=0;
	 fn=0;

	 for(i=0;i<n_vars-1;i++)
		for(j=i+1;j<n_vars;j++)
			if(net_local[i][j]==1 || net_local[j][i]==1)
				{if(the_true_net_local[i][j]==1 || the_true_net_local[j][i]==1)
					tp++;
				 else
					fp++;
				}
			else
				if(the_true_net_local[i][j]==1 || the_true_net_local[j][i]==1)
					fn++;

	 if(tp==0 && fp==0)
		printf(" p=%.2f",(double)(-1.0));
	 else
		printf(" p=%.2f",(double)tp/(double)(tp+fp));

	 if(tp==0 && fn==0)
		printf(" r=%.2f",(double)(-1.0));
	 else
		printf(" r=%.2f",(double)tp/(double)(tp+fn));

	 if((tp==0 && fp==0) || (tp==0 && fn==0))
		printf(" d=%.2f,",(double)(-1.0));
	 else
		printf(" d=%.2f,",sqrt(pow((double)1.0-(double)tp/(double)(tp+fp),(double)2.0)+pow((double)1.0-(double)tp/(double)(tp+fn),(double)2.0)));

	 printf("\n\n Inmoralities\n");
	 printf(" ------------");

	 printf("\n %d seconds,",(int)(clock()-clock_init)/(int)CLOCKS_PER_SEC);

	 tp=0;
	 fp=0;
	 fn=0;

	 for(i=0;i<n_vars-1;i++)
		for(j=i+1;j<n_vars;j++)
			for(k=0;k<n_vars;k++)
				if(net_local[i][k]==1 && net_local[k][i]==0 && net_local[j][k]==1 && net_local[k][j]==0 && net_local[i][j]==0 && net_local[j][i]==0)
					{if(the_true_net_local[i][k]==1 && the_true_net_local[k][i]==0 && the_true_net_local[j][k]==1 && the_true_net_local[k][j]==0 && the_true_net_local[i][j]==0 && the_true_net_local[j][i]==0)
						tp++;
					 else
						fp++;
					}
				else
					if(the_true_net_local[i][k]==1 && the_true_net_local[k][i]==0 && the_true_net_local[j][k]==1 && the_true_net_local[k][j]==0 && the_true_net_local[i][j]==0 && the_true_net_local[j][i]==0)
						fn++;

	 if(tp==0 && fp==0)
		printf(" p=%.2f",(double)(-1.0));
	 else
		printf(" p=%.2f",(double)tp/(double)(tp+fp));

	 if(tp==0 && fn==0)
		printf(" r=%.2f",(double)(-1.0));
	 else
		printf(" r=%.2f",(double)tp/(double)(tp+fn));

	 if((tp==0 && fp==0) || (tp==0 && fn==0))
		printf(" d=%.2f,",(double)(-1.0));
	 else
		printf(" d=%.2f,",sqrt(pow((double)1.0-(double)tp/(double)(tp+fp),(double)2.0)+pow((double)1.0-(double)tp/(double)(tp+fn),(double)2.0)));
	
	 for(i=0;i<n_vars;i++)
		delete [] the_true_net_local[i];
	 delete [] the_true_net_local;
	}

printf("\n\n Net\n");
printf(" ---");

printf("\n\n digraph G { size=\"7.50,10\" { ");

for(i=0;i<n_vars;i++)
	for(j=0;j<n_vars;j++)
		if(net_local[i][j]==1 || net_local[j][i]==1)
			{printf("%s ",names[i]);
			 break;
			}

printf(";");

for(i=0;i<n_vars;i++)
	for(j=0;j<n_vars;j++)
		if(net_local[i][j]==1 && net_local[j][i]==0)
			printf("\n %s -> %s;",names[i],names[j]);
		else
			if(net_local[i][j]==1 && net_local[j][i]==1 && i<j)
				printf("\n %s -> %s [arrowhead=none];",names[i],names[j]);

printf(" } }");

delete [] in_radius;
for(i=0;i<n_vars;i++)
	delete [] net_local[i];
delete [] net_local;
}
